﻿using System;
using System.Collections.Concurrent;
using System.Net;
using System.Net.WebSockets;
using WingServerCommon.Log;
using WingInterfaceLibrary.Enum;
using WingInterfaceLibrary.Request.Authentication;
using WingInterfaceLibrary.Interface;
using WingNotificationModule.Channel.WebSocket;
using System.Threading.Tasks;
using WingNotificationModule.Adapter;
using WingInterfaceLibrary.Notifications;
using System.Collections.Generic;
using System.Linq;
using WingInterfaceLibrary.Internal.Request;

namespace WingNotificationModule
{
    internal partial class NotificationServer
    {
        private readonly ConcurrentDictionary<string, WebScoketLeaf> _leaves = new ConcurrentDictionary<string, WebScoketLeaf>();
        private HttpListener _httpListener;
        private IAuthenticationService _authenticationService;
        private readonly string _host;
        private readonly NotificationQueueManager _notifcationQueueManager;

        public NotificationServer(string host, int maxQueueCount)
        {
            _host = host;
            _notifcationQueueManager = new NotificationQueueManager(maxQueueCount, SendMessage);
        }

        /// <summary>
        /// Start server
        /// </summary>
        public void Start(IAuthenticationService authenticationService)
        {
            _authenticationService = authenticationService;
            _httpListener = new();
            _httpListener.Prefixes.Add(_host);
            _httpListener.Start();
            _httpListener.BeginGetContext(new AsyncCallback(GetContextCallBackAsync), _httpListener);
            Logger.WriteLineInfo($"{_host} is listening.");
        }

        private void GetContextCallBackAsync(IAsyncResult ar)
        {
            HttpListener listener = (HttpListener)ar.AsyncState;
            HttpListenerContext context = listener.EndGetContext(ar);
            try
            {
                listener.BeginGetContext(new AsyncCallback(GetContextCallBackAsync), listener);
                var token = context?.Request?.QueryString["Token"]?.ToString();
                Logger.WriteLineInfo($"GetContextCallBackAsync token:{token}");
                HandleWebSocketAccept(context, token);
            }
            catch (Exception ex)
            {
                try
                {
                    Logger.WriteLineWarn($"GetContextCallBackAsync exception:{ex}");
                    context.Response.Abort();
                }
                catch (Exception exx)
                {
                    Logger.WriteLineWarn($"context.Response.Abort() exception:{exx}");
                }
            }
        }


        async void HandleWebSocketAccept(HttpListenerContext context, string token)
        {
            WebScoketLeaf leaf = null;
            try
            {
                WebSocketContext webSocketContext = await context.AcceptWebSocketAsync(subProtocol: null);
                var webSocket = webSocketContext.WebSocket;
                var state = webSocket.State;
                if (state == WebSocketState.Open)
                {
                    leaf = new WebScoketLeaf(webSocket, token);
                    //check token id valid
                    if (string.IsNullOrEmpty(token))
                    {
                        await leaf.SendAsync(new DisconnectNotification()
                        {
                        });
                        //scoketLeaf.Close(); //TODO if need to close leaf at server ?
                        return;
                    }

                    var result = await _authenticationService.ValidateTokenAsync(new ValidateTokenRequest() { Token = token });    //TODO do we need know why the token invalid ? 
                    if (result == null || result.Code != CustomerRpcCode.Ok)
                    {
                        await leaf.SendAsync(new DisconnectNotification()
                        {
                        });
                        return;
                    }

                    //cache web socket connection
                    _leaves.AddOrUpdate(token, (t) =>
                    {
                        Logger.WriteLineInfo($"Websocket connection with token {token} status:{state}, add leaf");
                        return leaf;
                    }, (t, exist) =>
                    {
                        Logger.WriteLineInfo($"Websocket connection with token {token} status:{state}, replace leaf");
                        exist.Close();
                        return leaf;
                    });
                    await leaf.SendAsync(new ConnectionNotification());
                }
                else
                {
                    Logger.WriteLineWarn($"Websocket connection with token {token} status:{state}");
                }
            }
            catch (Exception ex)
            {
                Logger.WriteLineError($"HandleWebSocketAccept with token {token} exception: {ex}");
            }
        }

        /// <summary>
        /// Send a message to client and return immediately after sending
        /// </summary>
        /// <param name="request">the message</param>
        /// <returns>bool</returns>
        /// <value>true/false</value>
        public async Task<bool> SendMessageAsync(SendNotificationRequest request)
        {
            var tokens = await _authenticationService.GetTokensWithClientIdAsync(new GetTokensWithClientIdRequest() { ClientId = request.ClientId });
            foreach (var token in tokens)
            {
                if (_leaves.TryGetValue(token.Code, out var leaf))
                {
                    await leaf.SendAsync(request.Message);
                    return true;
                }
                else
                {
                    Logger.WriteLineWarn($"Not found connection for token {token}");
                }
            }
            return false;
        }


        /// <summary>
        /// Send a message to client and return immediately after sending
        /// </summary>
        /// <param name="request">the message</param>
        /// <returns>bool</returns>
        /// <value>true/false</value>
        public async Task<bool> SendMessageAsync(IList<TokenDTO> tokens, SendNotificationRequest request)
        {
            foreach (var token in tokens)
            {
                if (_leaves.TryGetValue(token.Code, out var leaf))
                {
                    await leaf.SendAsync(request.Message);
                    return true;
                }
                else
                {
                    Logger.WriteLineWarn($"Not found connection for token {token}");
                }
            }
            return false;
        }




        /// <summary>
        /// Post a message to client and return immediately after add to sending queue
        /// </summary>
        /// <param name="request">the message</param>
        /// <returns>bool</returns>
        /// <value>true/false</value>
        public async Task<bool> PostMessageAsync(SendNotificationRequest request)
        {
            var tokens = await _authenticationService.GetTokensWithClientIdAsync(new GetTokensWithClientIdRequest() { ClientId = request.ClientId });
            foreach (var token in tokens)
            {
                Logger.WriteLineInfo($"NotificationService PostMessageAsync, account name:{token.AccountName}, token: {token.Code}, message:{request.Message}");
                _notifcationQueueManager.Post(string.Empty, new PostMessageParam(token.Code, request.Message));
            }
            return true;
        }


        /// <summary>
        /// Post a message to client and return immediately after add to sending queue
        /// </summary>
        /// <param name="request">the message</param>
        /// <returns>bool</returns>
        /// <value>true/false</value>
        public async Task<bool> PostMessageAsync(IList<TokenDTO> tokens, object messageData)
        {
            foreach (var token in tokens)
            {
                Logger.WriteLineInfo($"NotificationService PostMessageAsync, account name:{token.AccountName}, token: {token.Code}, message:{messageData}");
                _notifcationQueueManager.Post(string.Empty, new PostMessageParam(token.Code, messageData));
            }
            return true;
        }

        private bool SendMessage(PostMessageParam param)
        {
            try
            {
                var token = param.Token;
                if (_leaves.TryGetValue(token, out var leaf))
                {
                    leaf.Send(param.Message);
                }
                else
                {
                    Logger.WriteLineWarn($"Not found connection for token {token} when post message");
                }
            }
            catch (Exception ex)
            {
                Logger.WriteLineError($"post message ex:{ex}");
            }
            return true;
        }


        /// <summary>
        /// 开启独立队列
        /// </summary>
        /// <param name="module"></param>
        /// <returns></returns>
        public string OpenNotifyQueueAsync(string module)
        {
            var msgQueueId = module + "_" + Guid.NewGuid().ToString("N").ToUpper();
            var openResult = _notifcationQueueManager.TryAdd(msgQueueId);
            return openResult ? msgQueueId : "DefaultMessageQueue";//todo 如果add失败返回什么
        }


        /// <summary>
        /// Post a message to clients and return immediately after add to sending queue
        /// </summary>
        /// <param name="request">the message</param>
        /// <returns>bool</returns>
        /// <value>true/false</value>
        public async Task<bool> BroadcastMessageAsync(BroadcastNotificationRequest request)
        {
            var tokens = await _authenticationService.GetTokenWithClientIdsAsync(new GetTokenWithClientIdsRequest() { ClientIds = request.ClientIds.ToList() });
            foreach (var token in tokens)
            {
                Logger.WriteLineInfo($"NotificationService BroadcastMessageAsync, account name:{token.AccountName}, token: {token.Code}, msgQueueId:{request.MsgQueueId}, message:{request.Message}");
                _notifcationQueueManager.Post(request.MsgQueueId, new PostMessageParam(token.Code, request.Message));
            }
            return true;
        }


        /// <summary>
        /// Post a message to clients and return immediately after add to sending queue
        /// </summary>
        /// <param name="request">the message</param>
        /// <returns>bool</returns>
        /// <value>true/false</value>
        public async Task<bool> BroadcastMessageAsync(IList<TokenDTO> tokens, BroadcastNotificationRequest request)
        {
            foreach (var token in tokens)
            {
                Logger.WriteLineInfo($"NotificationService BroadcastMessageAsync, account name:{token.AccountName}, token: {token.Code}, msgQueueId:{request.MsgQueueId}, message:{request.Message}");
                _notifcationQueueManager.Post(request.MsgQueueId, new PostMessageParam(token.Code, request.Message));
            }
            return true;
        }
        /// <summary>
        /// 关闭独立队列
        /// </summary>
        /// <param name="module"></param>
        /// <returns></returns>
        public bool CloseNotifyQueueAsync(string queueId)
        {
            return _notifcationQueueManager.TryRemove(queueId);
        }

        /// <summary>
        /// Stop server
        /// </summary>
        public void Stop()
        {
            _httpListener.Stop();
        }


        internal class PostMessageParam
        {
            /// <summary>
            /// The token hash code 
            /// </summary>
            /// <value></value>
            public string Token { get; }

            /// <summary>
            /// The request message from caller
            /// </summary>
            /// <value></value>
            public object Message { get; }

            public PostMessageParam(string token, object message)
            {
                this.Token = token;
                this.Message = message;
            }
        }


        private class WebScoketLeaf
        {
            WebSocketIO _webSocketIO;
            public string ClientToken { get; }
            public WebScoketLeaf(WebSocket webSocket, string token)
            {
                ClientToken = token;
                _webSocketIO = new WebSocketIO(webSocket);
            }

            /// <summary>
            /// Send messge async
            /// </summary>        
            /// <param name="message">The message will be send to client</param>
            /// <returns></returns>
            public async Task SendAsync(object message)
            {
                if (_webSocketIO.IsDisconnected)
                {
                    throw new Exception("WebSocket is closed");
                }
                var adapter = new BufferAdapter(message);
                var buffer = adapter.GetMessageBuffer();
                await _webSocketIO.SendAsync(buffer, WebSocketMessageType.Binary);
            }


            /// <summary>
            /// Send messge
            /// </summary>        
            /// <param name="message">The message will be send to client</param>
            /// <returns></returns>
            public void Send(object message)
            {
                if (_webSocketIO.IsDisconnected)
                {
                    throw new Exception("WebSocket is closed");
                }
                var adapter = new BufferAdapter(message);
                var buffer = adapter.GetMessageBuffer();
                _webSocketIO.Send(buffer, WebSocketMessageType.Binary);
            }

            /// <summary>
            /// Close the web socket leaf connection
            /// </summary>
            public void Close()
            {
                _webSocketIO.Dispose();
            }
        }
    }
}